#!/usr/bin/env python3

# Nicola Heile
# nicola@heile.eu 
#
# Talk with Resol VBUS over LAN
# Copied some code parts from Github project:
# https://github.com/rellit/resol-vbus-python
# The origial project was designed for Network comunication,
# this project comunicates Via Serial Converter from 
# http://hobbyelektronik.org/w/index.php/VBus-Decoder

import serial
import time
import sys

# Load settings
try:
    import config
except:
    sys.exit("config.py not found!")


# Logs in onto the DeltaSol BS Plus over LAN. Also starts (and maintains) the
# actual stream of data.
def login(): # From Network Login - For serial only flush input Buffer and test connection
    try:
        dat = recv()

        return True
    except:
        return False


def load_data():
    if not config.simulation:
        buf = readstream()
        messages = splitmsg(buf)
        messages = selectDestination(messages, '0x0010') # 0x0010 0xFFFF DFA adressed
    else:
        # recorded message, with Temp1 = 12.2 , Temp2 = 9.8 , Drehzahl = 0
        messages = [[170, 16, 0, 123, 66, 16, 0, 1, 9, 24, 122, 0, 98, 0, 0, 35, 98, 0, 56, 34, 4, 63, 0, 127, 0, 0, 2, 126, 0, 127, 0, 0, 2, 126, 11, 1, 127, 127, 12, 105, 0, 0, 16, 3, 4, 104, 0, 0, 0, 0, 0, 127, 0, 0, 0, 0, 0, 127, 104, 0, 0, 0, 0, 23]]

    if config.debug:
        print("messages = " + str(messages))
    
    for msg in messages:
        if "PV1" == get_protocolversion(msg):
            if config.debug & False:
                print(format_message_pv1(msg))
            parse_values(msg)
        elif "PV2" == get_protocolversion(msg):
            if config.debug:
                print(format_message_pv2(msg))


# Receive 1024 bytes from stream
def recv():
    dat = ser.read(1)
    ret = int('0x' + dat.hex(),16)
    return ret


# Read Data until minimum 6 message is received
def readstream():
    data = [recv()]
    while data.count((0xAA)) < 6:
        data.append(recv())
    return data


#Split Messages on Sync Byte
def splitmsg(buf):
    x=[]
    buf=buf[buf.index(0xaa):]
    while buf.count(0xAA) > 1:
        x.append(buf[:buf.index(0xaa, 2)])
        buf = buf[buf.index(0xAA, 2) :]
    return x

def selectDestination(msg, dest):
    ret = []
    if config.debug: print("unfitered: msg = " + str(msg)) 
    for m in msg:
        if get_destination(m) == dest: 
            ret.append(m)
    if config.debug: print("fitered destinations " + str(dest) + ": msg = " + str(ret)) 
    return ret

# Format 1 byte as String
def format_byte(byte):
    if type(byte) is int:
        return hex(byte)[0:2] + '0' + hex(byte)[2:] if len(hex(byte)) < 4 else hex(byte)
    if type(byte) is list:
        x=[]
        for b in byte:
            x.append(format_byte(b))
        return x

# Extract protocol Version from msg
def get_protocolversion(msg):
    if hex(msg[5]) == '0x10': 
        if config.debug: print("Protocol Version 1") 
        return "PV1"
    #if hex(msg[5]) == '0x20': 
    #    if config.debug: print("Protocol Version 2")
    #    return "PV2"
    #if hex(msg[5]) == '0x30': 
    #    if config.debug: print("Protocol Version 3")
    #    return "PV3"
    if config.debug: print("Protocol Version UNKNOWN: " + str(msg[5]))
    return "UNKNOWN"


# Extract Destination from msg
def get_destination(msg):
    return format_byte(msg[2]) + format_byte(msg[1])[2:]


# Extract source from msg
def get_source(msg):
    return format_byte(msg[4]) + format_byte(msg[3])[2:]


# Extract command from msg
def get_command(msg):
    return format_byte(msg[5]) + format_byte(msg[6])[2:]


# Get count of frames in msg
def get_frame_count(msg):
    c = gb(msg, 8, 9
    )
    if config.debug: print("Frame Count: " + str(c))    
    return c - 1 #check last frame 


# Extract payload from msg
def get_payload(msg):
    payload = []
    for i in range(get_frame_count(msg)): # Get Frame Count
        frame = msg[10+(i*6):16+(i*6)]
        if config.debug: print("===\nProcessing Frame " + str(i))
        if checkCRC(frame) == 0:
            data = integrate_septett(frame)
            if config.debug: print("Processing ok Data = " + str(format_byte(data)))
            payload.append(data) # Add Values with integrated Septett
        else:
            if config.debug: print("Processing Failed, wrong CRC")
            payload.append(False)
    return payload

# Frames for xxx with source address 0x1122
#   Offset Frame  Description
#   0       0 0   Temperatur Sensor 1 (Bits 0-7)
#   1       0 1   Temperatur Sensor 1 (Bits 8-15)
#   2       0 2   Temperatur Sensor 2 (Bits 0-7)
#   3       0 3   Temperatur Sensor 2 (Bits 8-15)
#   4       1 0   Temperatur Sensor 3 (Bits 0-7)
#   5       1 1   Temperatur Sensor 3 (Bits 8-15)
#   6       1 2   Temperatur Sensor 4 (Bits 0-7)
#   7       1 3   Temperatur Sensor 4 (Bits 8-15)
#   8       2 0   Drehzahl 1 (Bits 0-7)
#   9       2 1   ?
#   10      2 2   ?
#   11      2 3   ?
#   12      3 0   ?
#   13      3 1   ?
#   14      3 2   ?
#   15      3 3   ?Fehlernummer (Bits 0-7)
#   16      4 0   ?Fehlermaske (Bits 0-7)
#   17      4 1   ?Fehlermaske (Bits 8-15)
#   18      4 2   ?Fehler-Info 1 (Bits 0-7)
#   19      4 3   ?Fehler-Info 2 (Bits 0-7)
#   20      5 0   ?Fehler-Info 3 (Bits 0-7)
#   21      5 1   ?Fehler-Info 4 (Bits 0-7)
def parse_values(msg): # Pick my Values
    frame = get_payload(msg)
    if frame[0] != False:
        TempSensor1 = gb(frame[0], 0, 2) / 10
        print("Temperatur1 = " + str(TempSensor1))
    if frame[0] != False:
        TempSensor2 = gb(frame[0], 2, 4) / 10
        print("Temperatur2 = " + str(TempSensor2))
    if frame[2] != False:
        Drezahl = gb(frame[2], 0, 1)
        print("Drehzahl Pumpe = " + str(Drezahl))

def format_message_pv1(msg):
    parsed = "PARSED: \n"
    parsed += "    ZIEL".ljust(15,'.')+": " + get_destination(msg) + "\n"
    parsed += "    QUELLE".ljust(15,'.')+": " + get_source(msg) + "\n"
    parsed += "    PROTOKOLL".ljust(15,'.')+": " + get_protocolversion(msg) + "\n"
    parsed += "    BEFEHL".ljust(15,'.')+": " + get_command(msg) + "\n"
    parsed += "    ANZ_FRAMES".ljust(15,'.')+": " + str(get_frame_count(msg)) + "\n"
    parsed += "    CHECKSUM".ljust(15,'.')+": " + format_byte(msg[9]) + "\n"
    for i in range(get_frame_count(msg)):
        integrated = integrate_septett(msg[10+(i*6):16+(i*6)])
        parsed += ("    NB"+str(i*4+1)).ljust(15,'.')+": " + format_byte(msg[10+(i*6)]) + " - " + format_byte(integrated[0]) + "\n"
        parsed += ("    NB"+str(i*4+2)).ljust(15,'.')+": " + format_byte(msg[11+(i*6)]) + " - " + format_byte(integrated[1]) + "\n"
        parsed += ("    NB"+str(i*4+3)).ljust(15,'.')+": " + format_byte(msg[12+(i*6)]) + " - " + format_byte(integrated[2]) + "\n"
        parsed += ("    NB"+str(i*4+4)).ljust(15,'.')+": " + format_byte(msg[13+(i*6)]) + " - " + format_byte(integrated[3]) + "\n"
        parsed += ("    SEPTETT"+str(i+1)).ljust(15,'.')+": " + format_byte(msg[14+(i*6)]) + "\n"
        parsed += ("    CHECKSUM"+str(i+1)).ljust(15,'.')+": " + format_byte(msg[15+(i*6)]) 
        if checkCRC(msg[10+(i*6):16+(i*6)]) == 0:
            parsed += (" --> ok\n")
        else:
            parsed += (" --> fail\n")
            
    #parsed += "    PAYLOAD".ljust(15,'.')+": " + (" ".join(format_byte(b) for b in get_payload(msg)))+"\n"
    return parsed

def integrate_septett(frame):
    try:
        if config.debug: print("Integrating Septet: " + str(format_byte(frame)))
        data = []
        septet = frame[4]
        for j in range(4):
            if septet & (1 << j):
                data.append((frame[j]) | 0x80)
            else:
                data.append(frame[j])
        if config.debug: print("Integrating Septet: " + str(format_byte(data)))
        return data
    except:
        print("Error integrting septet")
        return frame
    
# Gets the numerical value of a set of bytes (respect Two's complement by value Range)
def gb(data, begin=0, end=100):  # GetBytes
    wbg = sum([0xff << (i * 8) for i, b in enumerate(data[begin:end])])
    s = sum([b << (i * 8) for i, b in enumerate(data[begin:end])])
    if s >= wbg/2:
        s = -1 * (wbg - s)
    return s

def checkCRC(frame):
    CRC = 0x7F
    for b in frame:
        CRC = (CRC - b) & 0x7F
    
    if config.debug: print("Calculated CRC: " + str(format_byte(CRC)) + " (" + str(CRC) + ")" )
    return CRC
    

if __name__ == '__main__':
    if not config.simulation:
        ser = serial.Serial(config.port)
        #result = dict()
        login()
    
    if not config.simulation:
        while True:
            load_data()
    else:
       load_data() 

    try:
        ser.close()
    except:
        pass
    ser = None


